import openai
import markdown
from categories import *


class Book:
    """
    This class represents a book.

    Attributes:
        chapter_amount: The amount of chapters the book has.
        words_per_chapter: The amount of words per chapter.
        category: The category of the book.
        topic: The topic of the book.
        title: The title of the book.
        chapter_titles: The titles of the chapters.
        structure: The structure of the book.
    """

    def __init__(self, chapter_amount: int, words_per_chapter: int, topic: str, category: str):
        """
        This is the constructor for the Book class. It initializes the object with the given parameters.
        :param chapter_amount: The number of chapters in the book (int).
        :param words_per_chapter: The number of words per chapter (int).
        :param topic: The topic of the book (str).
        :param category: The category of the book (fiction, non-fiction, etc) (str).
        """

        # Define chapter amount
        self.chapter_amount = chapter_amount

        # Define words per chapter
        self.words_per_chapter = words_per_chapter

        # Define topic
        self.topic = topic

        # Define category
        self.category = category

        print('Initializing..')

        # Get book using `get_category` method
        self.book = self.get_category()

        # Set the title of the book
        self.title = self.get_title()

        # Set the chapter titles
        self.chapter_titles = self.get_chapter_titles()

        # Set the structure of the book
        self.structure = self.get_structure()

        self.packed = None

    @staticmethod
    def __get_edit(input_text, instruction, temperature: float = 0):
        """
        This is a private, static method for getting an edited version of the given input text.
        :param input_text: The text to be edited.
        :param instruction: The instructions for the edit.
        :param temperature: The temperature of the edit (defaults to 0).
        :return: The edited text.
        """

        # Define the engine to use and returning the edited version
        return openai.Edit.create(
            model="text-davinci-edit-001",
            input=input_text,
            instruction=instruction,
            temperature=temperature).choices[0].text

    def get_category(self):

        # Get the category
        self.category = self.category

        # Return the category class
        return globals()[self.category](self.chapter_amount, self.words_per_chapter, self.topic)

    def get_title(self):
        """
        This method returns the title of the book.
        """

        # Get the title of the book and delete the blank lines
        title = self.book.get_title().replace('\n', '')

        # Return the title
        return title

    def get_chapter_titles(self):
        """
        This method returns the chapter titles of the book.
        """

        # Define the chapter titles as the chapter titles of the book
        chapters = self.book.get_chapters(self.title).split('\n')

        # Remove the blank lines
        chapters = [chapter for chapter in chapters if chapter != '']

        # Return the chapter titles
        return chapters

    def get_structure(self):
        """
        This method returns the structure of the book.
        """

        # Get the structure of the book and splitting it by the new lines
        structure = self.book.get_structure(self.title, self.chapter_titles).split('\n')

        # Remove the blank lines
        structure = [chapter for chapter in structure if chapter != '']

        # Create a list of chapters
        chapter_list = []

        # Create a list of paragraphs
        chapter = []

        # Iterate over the lines
        for line in structure:

            # If the line starts with 'Chapter'
            if line.lower().startswith('chapter'):
                # Add the chapter to the list of chapters
                chapter_list.append(chapter)

                # Create a new list of paragraphs
                chapter = []

            # Add the line to the list of paragraphs
            chapter.append(line)

        # Add the last chapter to the list of chapters
        chapter_list.append(chapter)

        # Remove the first chapter
        chapter_list = chapter_list[1:]

        # Remove the first paragraph of each chapter
        chapter_list = [chapter[1:] for chapter in chapter_list]

        # Create a new list of chapters
        chapter_list_new = []

        # Iterate over the chapters
        for chapter in chapter_list:

            # Create a list of paragraphs
            paragraph_list = []

            # Iterate over the paragraphs
            for paragraph in chapter:

                # Get the title of the paragraph
                title = paragraph.split('---')[0]

                # Get the word count of the paragraph
                word_count = paragraph.split('---')[1]

                # If the word count contains the word 'words'
                if 'words' in word_count.lower():
                    # Remove the word 'words'
                    word_count = word_count.split(' ')[0]

                # Add the paragraph to the list of paragraphs
                paragraph_list.append({'title': title, 'word_count': word_count})

            # Add the list of paragraphs to the list of chapters
            chapter_list_new.append(paragraph_list)

        # Return the list of chapters
        return chapter_list_new

    def get_paragraph(self, chapter_index, paragraph_index):
        """
        This method returns the paragraph at the given index in the given chapter.
        :param paragraph_index: The index of the paragraph.
        :param chapter_index: The index of the chapter.
        :return: The paragraph.
        """

        # Get the paragraph at the given index in the given chapter
        return self.book.get_paragraph(self.title, self.chapter_titles, self.structure, paragraph_index, chapter_index)

    def get_chapter(self, chapter_index):
        """
        This method returns the chapter at the given index.
        :param chapter_index: The index of the chapter.
        :return: The chapter.
        """

        # Define empty list for the paragraphs
        chapter = []

        # Iterate over the paragraphs in the chapter
        for i in range(len(self.structure[chapter_index])):

            # Add the paragraph to the list of paragraphs
            chapter.append(self.get_paragraph(chapter_index, i))

        # Return the chapter
        return chapter

    def get_content(self):
        """
        This method returns the book.
        :return: The book.
        """

        # Define empty list for the content of the chapters
        content = []

        # Iterate over the chapters
        for i in range(len(self.structure)):

            # Add the content of the chapter to the list
            content.append(self.get_chapter(i))

        # Return the content
        return content

    def generate(self):
        """
        This method generates the book.
        """

        # Get content
        content = self.get_content()

        # Pack book
        book = {'title': self.title, 'chapters': content, 'chapter_titles': self.chapter_titles, 'structure': self.structure}

        # Return the book
        self.packed = book

    def get_md(self):
        """
        This method returns the book in Markdown format.
        """

        # Check if variable packed is exists
        if self.packed is None:
            print('Book not generated yet. Generating book...')
            self.generate()

        # Get the book
        book = self.packed

        # Define the variables
        content = book['chapters']

        # Build the book

        # Add the title of the book
        book = '# ' + self.title + '\n\n\n'

        # Add the chapter titles and content
        for i in range(len(content)):

            # Add the chapter title in md format
            book += '\n\n## ' + self.chapter_titles[i] + '\n'

            # Iterate through the paragraphs
            for paragraph in content[i]:

                # Check if paragraph ends with a blank line
                if paragraph.endswith('\n'):

                    # Add the paragraph
                    book += paragraph
                else:

                    # Add the paragraph with a blank line
                    book += paragraph + '\n'

        # Return the combined book.
        return book

    def get_html(self):
        """
        This method returns the book in HTML format.
        """

        # Getting Markdown element
        md = markdown.markdown(self.get_md())

        # Return the HTML
        return md



import os


def draw_data_structure(data, indent=0):
    for key, value in data.items() if isinstance(data, dict) else enumerate(data):
        if isinstance(value, (dict, list)):
            print(' ' * indent + '\n' + str(key))
            draw_data_structure(value, indent + 2)
        else:
            print(' ' * (indent + 2) + str(value))


def get_python_files(path):
    return [file[:-3] for file in os.listdir(path) if not file.startswith('__') and file.endswith('.py')]


def get_categories():
    # Get all the files in the directory 'categories'
    categories = get_python_files('categories')

    # Check if the category is in the directory
    special_names = {'selfimprovement': 'SelfImprovement'}

    # Return all categories with the first letter capitalized, except for the special names, which are returned, how they are defined in special_names
    return [category.capitalize() if category not in special_names else special_names[category] for category in categories]



from setuptools import setup, find_packages
import codecs
import os

VERSION = '0.6.2.1'
DESCRIPTION = 'Generating books using OpenAI\'s GPT-3 API.'
# Set README as long description
with codecs.open('README.md', 'r', 'utf-8') as f:
    LONG_DESCRIPTION = f.read()

# Setting up
setup(
    name="bookgpt",
    version=VERSION,
    author="Mika Vehns",
    author_email="<mika.vehns@outlook.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=['openai', 'markdown', 'retrying'],
    keywords=['python', 'gpt-3', 'openai', 'book', 'generator', "writer"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ]
)



import openai
import retrying


class Science:
    """
    This class is used to generate a self-help book.
    """

    def __init__(self, chapter_amount: int, words_per_chapter: int, topic: str):
        """
        Initialize the class.
        :param chapter_amount: The amount of chapters in the book.
        :param words_per_chapter: The amount of words per chapter.
        :param topic: The topic of the book.
        """

        self.chapter_amount = chapter_amount
        self.words_per_chapter = words_per_chapter
        self.topic = topic

    @staticmethod
    @retrying.retry(stop_max_attempt_number=5, wait_fixed=1000, retry_on_exception=lambda e: isinstance(e, openai.error.ServiceUnavailableError or openai.error.RateLimitError))
    def get_response(prompt: str):
        """
        Gets a response from the API.
        :param prompt: The prompt to send to the API.
        :return: The response from the API.
        """

        return openai.Completion.create(
            engine="text-davinci-003",
            prompt=prompt,
            temperature=0.7,
            max_tokens=1500,
            top_p=1,
            frequency_penalty=0,
            presence_penalty=0
        ).choices[0].text

    def get_title(self):
        """
        Gets the title of the book.
        :return: The title of the book.
        """

        return self.get_response(f"Generate a title for a science book on {self.topic}. "
                                 f"The title should be engaging and memorable, and should "
                                 f"accurately reflect the content and purpose of the book. "
                                 f"The book will contain scientific information, experiments, "
                                 f"and research to help readers understand and learn about the topic. "
                                 f"The title should be informative and educational, and should inspire "
                                 f"readers to further explore the subject.")

    def get_chapters(self, title: str):
        """
        Gets the chapters of the book.
        :param title: The title of the book.
        :return: The chapters of the book.
        """

        return self.get_response(
            f"Generate a list of the size {self.chapter_amount} of chapter titles for a science book on the topic {self.topic}, called {title}. "
            f"Each chapter should cover a specific aspect of the topic and should be structured as a "
            f"series of lessons or explanations that the reader can follow to gain a deeper understanding of the subject. "
            f"The chapter titles should be descriptive and should clearly convey the main focus of each chapter. "
            f"The chapters should be informative and educational, and should encourage the reader to think critically about the topic.")

    def get_structure(self, title: str, chapters: list[str]):
        """
        Gets the structure of the book.
        :param title: The title of the book.
        :param chapters: The chapters of the book.
        :return: The structure of the book.
        """

        return self.get_response(f"Generate a structure plan for a science book called {title}. "
                                 f"The book should contain {self.chapter_amount} chapters, with the following titles: "
                                 f"{','.join(chapters)}. Each chapter should cover a specific topic in science and "
                                 f"be structured as a series of lessons or sections that explain the key concepts and ideas. "
                                 f"The chapters should include examples, illustrations, and case studies to help the reader "
                                 f"understand and apply the concepts. The book should be informative and accurate, "
                                 f"and should encourage the reader to think critically about the topic."
                                 f"\n\nFor each chapter, create a list of paragraph titles and corresponding recommended"
                                 f" word counts in the following format: 'paragraph_title---word_amount.' "
                                 f"The section titles should not include the word 'paragraph' or a number. "
                                 f"The total recommended word count for all sections in each chapter should add up to {self.words_per_chapter} words. "
                                 f"In order to prevent any individual section from being too long, try to divide the "
                                 f"content into multiple sections, each with a recommended word count.")

    def get_paragraph(self, title: str, chapters: list[str], paragraphs: list[list[dict[str, str]]],
                      paragraph_index: int, chapter_index: int):
        """
        Gets a paragraph of the book.
        :param title: The title of the book.
        :param chapters: The chapters of the book.
        :param paragraphs: The paragraphs of the book.
        :param paragraph_index: The index of the paragraph.
        :param chapter_index: The index of the chapter.
        :return: The paragraph of the book.
        """

        paragraphs = paragraphs[chapter_index]
        titles = '\n'.join([paragraph["title"] + ' - ' + paragraph["word_count"] for paragraph in paragraphs])
        paragraph = paragraphs[paragraph_index]
        return self.get_response(
            f"Write the content for Chapter {chapter_index + 1} of a science book called {title}. The chapter is called {chapters[chapter_index]}, "
            f"and should focus on discussing scientific theories, principles, and research findings related to the topic. "
            f"The chapter should include a clear explanation of the key concepts, as well as supporting evidence and examples to illustrate their importance. "
            f"The chapter should also provide practical applications of the concepts, and may include hands-on activities or exercises for readers to try. "
            f"The chapter should be written in a clear and concise style, and should be well-organized and easy to follow. "
            f"\n\nThe chapter contains the following paragraphs:\n{titles}\nWrite the content for paragraph {paragraph_index + 1}, "
            f"with the title '{paragraph['title']}.' The paragraph should have the recommended word count of {paragraph['word_count']} words. "
            f"In order to effectively convey the main ideas and concepts of the paragraph, be sure to include relevant data, "
            f"experiments, and other scientific evidence. Use a logical and scientific writing style that will keep the reader engaged and informed. "
            f"The paragraph should be well-structured and coherent, with a clear introduction, body, and conclusion. Do not write a title.")
